#include	"Controller.h"

/*
 * Controller.cpp (c) Noah Roberts 2003-03-02
 */

#include	"Engine.h"
#include	"Lawyer.h"
#include	"Board.h"
#include	"Interface.h"
#include	"Move.h"
#include	"Evaluator.h"
#include	"Command.h"
#include	"Timer.h"

#include	<cstdio>

#ifdef WIN32
# include	<windows.h>
#else
# include	<unistd.h>
#endif

#include	<strstream>

using namespace std;

Controller::Controller(int argc, char **argv)
{
  board = new Board();
  lawyer = new Lawyer(board);
  board->addObserver(lawyer);
  interfce = new Interface(this);
  engine = new Engine(board, lawyer, interfce);
  command = new Command(interfce, board, engine, lawyer, this);
  computerColor = BLUE;
  cxboard = false;
  Options::defaultOptions()->addObserver(this);
  Options::defaultOptions()->decipherCommandArgs(argc, argv);
}
Controller::~Controller()
{
  delete board;
  delete lawyer;
  delete engine;
  delete interfce;
}

bool Controller::isMove(Move &move)
{
  if (move.origin() == 0 && move.destination() == 0)
    {
      return false;
    }
  return true;
}

void Controller::run()
{
  bool cont = true;
  Move move;
  string commandText;
  interfce->printBoard(*board);
  interfce->printPrompt();
  bool userTimerStarted = false;
  while (cont) // Until the user wants to quit...
    {
      // Check to see if the game is over based on return of move generation...
      if (!board->gameOver())
        {
          color x;
          list<Move> moveList;
          lawyer->generateMoves(moveList, true);
          if (moveList.empty())
            {
              string reason;
              if (lawyer->inCheck())
                reason = string(board->sideToMove() == RED ? "Blue":"Red") +  " mates";
              else
                reason = string(board->sideToMove() == RED ? "Red":"Blue") + " stalemated";
              (*interfce) << (board->sideToMove() == RED ? "0-1 {":"1-0 {") << reason << "}\n";
              board->gameOver(true);
            }
          else if (lawyer->drawn()) // Game is drawn if no pieces can cross the river.
            {
              board->gameOver(true);
              (*interfce) << "1/2-1/2 {" << lawyer->getMessage() << "}\n";
            }
          // Timer controls...
          else if (Timer::redTimer()->active() &&
                   Timer::redTimer()->surpasedTime())
            {
              board->gameOver(true);
              (*interfce) << "0-1 {Red runs out of time}\n";
            }
          else if (Timer::blueTimer()->active() &&
                   Timer::blueTimer()->surpasedTime())
            {
              board->gameOver(true);
              (*interfce) << "1-0 {Blue runs out of time}\n";
            }
          else if ((x = (color)lawyer->gameWonByPCheck()) != NOCOLOR)
            {
              board->gameOver(true);
              string result(board->sideToMove() == RED ? "1-0":"0-1");
              (*interfce) << result << " {Perpetual Check}\n";
              //cerr << "Color that won: " << (x == RED ? "RED":"BLUE") << endl;
            }
          else if ((x = (color)lawyer->gameWonByChase()) != NOCOLOR)
            {
              board->gameOver(true);
              string result(board->sideToMove() == RED ? "1-0":"0-1");
              (*interfce) << result << " {Perpetual Chase}\n";
              // cerr << "Color that won: " << (x == RED ? "RED":"BLUE") << endl;
            }
        }
      // Deal with user input first...
      if (interfce->readReady())
        {
          commandText = interfce->readLine();

          // Move...
          move = Move(commandText);
          if (isMove(move))
            {
              if (board->colorAt(move.origin()) == board->sideToMove() && (computerColor == NOCOLOR || board->sideToMove() != computerColor))
                {
                  if (lawyer->legalMove(move))
                    {
                      Timer::timerForColor(board->sideToMove())->stopTimer();
                      Timer::timerForColor(board->sideToMove())->moveMade();
                      //userTimerStarted = false;
                      board->makeMove(move);
                      interfce->printBoard(*board);
                      if (!cxboard)
                        interfce->printMove(move);
                      Timer::timerForColor(board->sideToMove())->startTimer();
                    }
                  else
                    {
                      (*interfce) << "IllegalMove";
                      if (lawyer->getMessage() != "")
                        (*interfce) << " (" << lawyer->getMessage() << ")";
                      (*interfce) << ": " << (move).getText() << "\n";
                    }
                }
              else
                {
                  (*interfce) << "IllegalMove (Out of turn): " << move.getText() << "\n";
                }
            }
          // Option...
          else if (Options::defaultOptions()->isOption(commandText));
          // Command...
          else if (command->isCommand(commandText));
          // Unknown...
          else (*interfce) << "Error (unknown command): " << commandText << "\n";

          //interfce->printBoard();
          interfce->printPrompt();
        }
      else if (board->sideToMove() == computerColor && !board->gameOver())
        {
          if (engine->doneThinking())
            {
              move = engine->getMove();
              Move x = Move();
              if (!(move == x))
                {
                  Timer::timerForColor(board->sideToMove())->stopTimer();
                  Timer::timerForColor(board->sideToMove())->moveMade();
                  board->makeMove(move);
                  interfce->printBoard(*board);
                  interfce->printMove(move);
                  interfce->printPrompt();
                  Timer::timerForColor(board->sideToMove())->startTimer();
                }
            }
          else
            engine->think();
        }
#ifdef WIN32
      Sleep(100);
#else
      usleep(100);
#endif
    }
}

void Controller::optionChanged(string whatOption) // Pay attention to configuration changes...
{
  if (whatOption == "computerColor")
    {
      string choice = Options::defaultOptions()->getValue(whatOption);
      if (choice == "r") computerColor = RED;
      else if (choice == "b") computerColor = BLUE;
      else computerColor = NOCOLOR;
    }
  else if (whatOption == "interface")
    {
      if (Options::defaultOptions()->getValue(whatOption) == "cxboard")
        cxboard = true;
      else
        cxboard = false;
    }
  else if (whatOption == "st")
    {
      string s = Options::defaultOptions()->getValue(whatOption);
      istrstream in(s.c_str(), s.size());
      int x;
      in >> x;
      if (in)
        {
          Timer::setTimers(1, x, 0);
          Timer::timerForColor(board->sideToMove())->startTimer();
        }
    }
}
