#include	"Interface.h"
#include	"Controller.h"

#ifdef WIN32
#include	<windows.h>
#include	<conio.h>
#else
#include	<sys/select.h>
#include	<unistd.h>
#endif

/*
 * Interface.cpp (c) Noah Roberts 2003-03-01
 */

using namespace std;
Interface::Interface(Controller *c)
{
  control = c;
  state = new UserState(c);
  cxboard = false;
  Options::defaultOptions()->addObserver(this);
}
Interface::~Interface()
{
  // Delete our state object.
  if (state != NULL) delete state;
}

void Interface::setState(string whichState) // Changes the behavior of class...
{
  if (whichState == "cxboard") // Talk to GUI
    {
      if (state != NULL) delete state;
      state = new CXBoardState(control);
      cxboard = true;
    }
  else if (whichState == "user") // Talk to user...
    {
      if (state != NULL) delete state;
      state = new UserState(control);
      cxboard = false;
    }
}
void Interface::optionChanged(string whatOption) // Pay attention to configuration changes.
{
  if (whatOption == "interface")
    setState(Options::defaultOptions()->getValue(whatOption));
}

#ifdef WIN32
bool Interface::readReady()
{
  static bool init = false, pipe = false;
  static HANDLE in;
  bool ready = false;

  DWORD dw;
  if (!init)
    {
      init = true;
      in = GetStdHandle(STD_INPUT_HANDLE);
      pipe = !GetConsoleMode(in, &dw); // So we know how we should poll...
      if (!pipe)
        {
          SetConsoleMode(in, dw & ~(ENABLE_MOUSE_INPUT|ENABLE_WINDOW_INPUT));
          FlushConsoleInputBuffer(in);
        }
    }
  /*
   * Must deal with MS brain deadedness now.
   */
  if (pipe) // Connected through stdin to another program
    {
      if (!PeekNamedPipe(in, NULL, 0, NULL, &dw, NULL))
        return 1;
      return dw;
    }
  else     // Normal console mode.
    {
      GetNumberOfConsoleInputEvents(in, &dw);
      return dw <= 1 ? 0: dw;
    }
  return ready;
}
#else
bool Interface::readReady()
{
  fd_set	fds; // the file descriptors to check
  struct timeval tv; // how long to wait
  int sel_val;       // select() return value

  FD_ZERO(&fds);
  FD_SET(0, &fds); // stdin
  tv.tv_sec = 0;   // don't wait at all...
  tv.tv_usec = 0;

  sel_val = select(1, &fds, NULL, NULL, &tv);

  if (sel_val) return true;
  else return false;
}
#endif /* WIN32 */

string Interface::readLine()
{
  string final;

  int c;
  while (cin && (c = cin.get()) != '\n') // Read up to '\n' and append it to final.
    {
      final += c;
    }

  return final;
}

void UserState::printMove(Move &theMove) // Display a move.
{
  Board *x = control->getBoard();
  if (x->colorAt(theMove.destination()) == RED)
    cout << "    " << theMove << " ..." << endl;
  else cout << "    ... " << theMove << endl << endl;
  cout.flush();
}

CXBoardState::CXBoardState(Controller *c)
{
#if BUF_STD_COMP
  cin.rdbuf()->pubsetbuf(NULL, 0);
#else
  cin.rdbuf()->setbuf(NULL, 0);
#endif
}

CXBoardState::~CXBoardState()
{
#if BUF_STD_COMP
  cin.rdbuf()->pubsetbuf(new char[512], 512);
#else
  cin.rdbuf()->setbuf(new char[512], 512);
#endif
}
