#ifndef	__INTERFACE_H__
#define	__INTERFACE_H__

#include	<iostream>
#include	<string>
#include	"Options.h"
#include	"Move.h"
#include	"Board.h"

class Controller;
/*
 * Interface.h (c) Noah Roberts 2003-02-27
 * Inteface class does all the I/O for the chess engine.
 * Uses the State pattern.
 */

class InterfaceState // Generic state
{
 protected:
  Controller *control;
 public:
  virtual void printMove(Move &theMove) {}
  virtual void printPrompt() {}
  virtual void printBoard(Board &theBoard) {}
  virtual void printText(std::string text) { std::cout << text; std::cout.flush();}
  virtual void printText(char c) { std::cout << c; std::cout.flush();}
  virtual void printInt(int x) { std::cout << (int)x; std::cout.flush();}
};

class Interface : public OptionsObserver
{
 protected:
  InterfaceState	*state;
  Controller		*control;
  bool cxboard;

  void setState(std::string which);
 public:
  Interface(Controller *c);
  ~Interface();
  bool readReady();
  void setController(Controller *c) { control = c; }
  std::string readLine();
  void printMove(Move &theMove) { state->printMove(theMove); }
  void printBoard(Board &theBoard) { state->printBoard(theBoard); }
  void printPrompt() { state->printPrompt(); }

  void optionChanged(std::string whatOption);

  Interface &operator<<(std::string text) { state->printText(text); return *this;}
  Interface &operator<<(char ch) { state->printText(ch); return *this;}
  Interface &operator<<(int x) { state->printInt(x); return *this; }
};

class UserState : public InterfaceState // State class for talking to a user.
{
 public:
  UserState(Controller* c) { control = c; }
  void printMove(Move &theMove);
  void printBoard(Board &theBoard) { std::cout << std::endl << theBoard << std::endl; std::cout.flush(); }
  void printPrompt() { std::cout << "Enter move or command: "; std::cout.flush();}
};

class CXBoardState : public InterfaceState // State class for talking to a program.
{
 public:
  CXBoardState(Controller *c);
  ~CXBoardState();
  void printMove(Move &theMove)
      { std::cout << "move " << theMove << std::endl; std::cout.flush();
        std::cerr << "engine move: " << theMove << std::endl; }
};

#endif /* __INTERFACE_H__ */
